import click
import secrets
import string

from . import util


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
ALPHABET = dict(
    nopunctuation=string.ascii_letters + string.digits,
    digits=string.digits,
    small=string.ascii_letters + string.digits + '.!?,;@#%&*-+_',
    full=string.ascii_letters + string.digits + '.!?,;:@#%&*-+=/|^_~()<>[]{}',
)


@click.command(options_metavar='<options>', context_settings=CONTEXT_SETTINGS)
@click.argument('size', default=16, type=click.IntRange(10, None), metavar='<size>')
@click.option('-s', '--small', is_flag=True, help='Small set of valid punctuation symbols for password.')
@click.option('-d', '--digits', is_flag=True, help='Only digits for password.')
@click.option('-n', '--nopunctuation', is_flag=True, help="Don't add punctuation symbols for password.")
@click.option('-b', '--batch', 'qty', default=1, type=click.IntRange(1, None), metavar='<qty>', help='To generate more than one password.')
@click.version_option(None, '-v', '--version', message="version %(version)s")
@util.doc(
    f"""Make <qty> random passwords for given <size> and optionally chosen alphabet type.

    \b
    size:               Password length (min: 10).
    alphabet type:
        default         {ALPHABET['full']}
        small           {ALPHABET['small']}
        nopunctuation   {ALPHABET['nopunctuation']}
        digits          {ALPHABET['digits']}
    """
)
def run(size, qty, **kwargs):
    for _ in range(qty):
        if kwargs['digits']:
            password = ''.join(secrets.choice(ALPHABET['digits']) for i in range(size))
            print(password)
        else:
            if kwargs['nopunctuation']:
                alphabet = ALPHABET['nopunctuation']
            elif kwargs['small']:
                alphabet = ALPHABET['small']
            else:
                alphabet = ALPHABET['full']

            while True:
                password = ''.join(secrets.choice(alphabet) for i in range(size))

                if (
                    not starts_with_punctuation(password) and
                    (has_punctuation(password) or kwargs['nopunctuation']) and
                    has_lowercase(password, 3) and
                    has_uppercase(password, 3) and
                    has_digits(password, 3)
                ):
                    print(password)
                    break


def starts_with_punctuation(password):
    return password[0] in string.punctuation


def has_punctuation(password):
    return any(c in string.punctuation for c in password)


def has_uppercase(password, times):
    return sum(c.isupper() for c in password) >= times


def has_lowercase(password, times):
    return sum(c.islower() for c in password) >= times


def has_digits(password, times):
    return sum(c.isdigit() for c in password) >= times


if __name__ == '__main__':
    run()



import os
import inspect


def doc(arg):
    """Docstring decorator.

    arg:    Docstring text or object.
    """
    def decorator(func):
        if type(arg) is str:
            func.__doc__ = arg
        elif inspect.isclass(arg):
            func.__doc__ = arg.__doc__
        else:
            func.__doc__ = None

        return func
    return decorator



import itertools
import pytest
import re
import string

from pyssword import pyssword as m


@pytest.mark.parametrize('option', ['-h', '--help'])
def test_run_help(cli_invoker, option):
    result = cli_invoker(m.run, option)
    assert result.exit_code == 0
    assert result.output.startswith('Usage: run <options> <size>')


@pytest.mark.parametrize('option', ['-v', '--version'])
def test_run_version(cli_invoker, option):
    result = cli_invoker(m.run, option)
    assert result.exit_code == 0
    assert re.search(r'version (\d\.?){3}(\.dev\d+)?$', result.output.strip(), re.IGNORECASE)


@pytest.mark.parametrize('size', [
    *[pytest.param(i, marks=pytest.mark.xfail) for i in range(0, 10)],
    *range(10, 21),
])
def test_run_size(cli_invoker, size):
    result = cli_invoker(m.run, f'{size}')
    assert result.exit_code == 0
    assert len(result.output.strip()) == size


@pytest.mark.parametrize('option,alphabet_type', itertools.chain(
    # options with full punctuation
    list(itertools.product(
        [''],
        ['full']
    )),
    # options with digits
    list(itertools.product(
        ['-d', '--digits'],
        ['digits']
    )),
    # options with small punctuation
    list(itertools.product(
        ['-s', '--small'],
        ['small']
    )),
    # options with nopunctuation
    list(itertools.product(
        ['-n', '--nopunctuation'],
        ['nopunctuation']
    )),
    # options with small and nopunctuation
    list(itertools.product(
        itertools.product(
            ['-s', '--small'],
            ['-n', '--nopunctuation']
        ),
        ['nopunctuation']
    )),
    # options with digits, small and nopunctuation
    list(itertools.product(
        itertools.product(
            ['-d', '--digits'],
            ['-s', '--small'],
            ['-n', '--nopunctuation']
        ),
        ['digits']
    )),
))
@pytest.mark.parametrize('size', range(10, 30))
def test_run_mixed_options(cli_invoker, option, alphabet_type, size):
    alphabet          = m.ALPHABET[alphabet_type]
    negative_alphabet = ''.join([c for c in string.printable if c not in alphabet])

    option = ' '.join(option) if isinstance(option, (list, tuple, set)) else option
    result = cli_invoker(m.run, f'{option} {size}')
    output = result.output.strip()

    assert result.exit_code == 0
    assert len(output) == size

    assert not any(c not in alphabet for c in output), f"Invalid punctuations --> {tuple(c for c in output if c not in alphabet)} in {output}"
    assert not any(c in negative_alphabet for c in output), f"Invalid punctuations --> {tuple(c for c in output if c in negative_alphabet)} in {output}"


@pytest.mark.parametrize('option', ['-b', '--batch'])
@pytest.mark.parametrize('qty', [
    *[pytest.param(i, marks=pytest.mark.xfail) for i in range(-2, 1)],
    *range(1, 11),
])
@pytest.mark.parametrize('size', range(10, 30))
def test_run_batch(cli_invoker, option, qty, size):
    result = cli_invoker(m.run, f'{option} {qty} {size}')
    output = result.output.split()

    assert result.exit_code == 0
    assert len(output) == qty

    for item in output:
        assert len(item.strip()) == size
